@@ -1,13 +1,13 @@
 /*
- * Copyright 2003-2007 Gentoo Foundation
+ * Copyright 2003-2012 Gentoo Foundation
  * Distributed under the terms of the GNU General Public License v2
- * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.244 2012/04/29 06:21:36 vapier Exp $
+ * $Header: /var/cvsroot/gentoo-projects/pax-utils/scanelf.c,v 1.250 2012/11/04 08:25:41 vapier Exp $
  *
- * Copyright 2003-2007 Ned Ludd        - <solar@gentoo.org>
- * Copyright 2004-2007 Mike Frysinger  - <vapier@gentoo.org>
+ * Copyright 2003-2012 Ned Ludd        - <solar@gentoo.org>
+ * Copyright 2004-2012 Mike Frysinger  - <vapier@gentoo.org>
  */
 
-static const char rcsid[] = "$Id: scanelf.c,v 1.244 2012/04/29 06:21:36 vapier Exp $";
+static const char rcsid[] = "$Id: scanelf.c,v 1.250 2012/11/04 08:25:41 vapier Exp $";
 const char argv0[] = "scanelf";
 
 #include "paxinc.h"
@@ -46,6 +46,8 @@ static char be_verbose = 0;
 static char be_wewy_wewy_quiet = 0;
 static char be_semi_verbose = 0;
 static char *find_sym = NULL;
+static array_t _find_sym_arr = array_init_decl, *find_sym_arr = &_find_sym_arr;
+static array_t _find_sym_regex_arr = array_init_decl, *find_sym_regex_arr = &_find_sym_regex_arr;
 static char *find_lib = NULL;
 static array_t _find_lib_arr = array_init_decl, *find_lib_arr = &_find_lib_arr;
 static char *find_section = NULL;
@@ -118,31 +120,6 @@ static const char *root_rel_path(const char *path)
 	return path;
 }
 
-/* 1 on failure. 0 otherwise */
-static int rematch(const char *regex, const char *match, int cflags)
-{
-	regex_t preg;
-	int ret;
-
-	if ((match == NULL) || (regex == NULL))
-		return EXIT_FAILURE;
-
-	if ((ret = regcomp(&preg, regex, cflags))) {
-		char err[256];
-
-		if (regerror(ret, &preg, err, sizeof(err)))
-			fprintf(stderr, "regcomp failed: %s", err);
-		else
-			fprintf(stderr, "regcomp failed");
-
-		return EXIT_FAILURE;
-	}
-	ret = regexec(&preg, match, 0, NULL, 0);
-	regfree(&preg);
-
-	return ret;
-}
-
 /* sub-funcs for scanelf_fileat() */
 static void scanelf_file_get_symtabs(elfobj *elf, void **sym, void **str)
 {
@@ -1176,27 +1153,14 @@ static void
 scanelf_match_symname(elfobj *elf, char *found_sym, char **ret, size_t *ret_len, const char *symname,
 	unsigned int stt, unsigned int stb, unsigned int shn, unsigned long size)
 {
-	char *this_sym, *next_sym, saved = saved;
+	const char *this_sym;
+	size_t n;
 
-	/* allow the user to specify a comma delimited list of symbols to search for */
-	next_sym = NULL;
-	do {
+	array_for_each(find_sym_arr, n, this_sym) {
 		bool inc_notype, inc_object, inc_func, inc_file,
 		     inc_local, inc_global, inc_weak,
 		     inc_def, inc_undef, inc_abs, inc_common;
 
-		if (next_sym) {
-			next_sym[-1] = saved;
-			this_sym = next_sym;
-		} else
-			this_sym = find_sym;
-		if ((next_sym = strchr(this_sym, ','))) {
-			/* make parsing easier by killing the comma temporarily */
-			saved = *next_sym;
-			*next_sym = '\0';
-			next_sym += 1;
-		}
-
 		/* symbol selection! */
 		inc_notype = inc_object = inc_func = inc_file = \
 		inc_local = inc_global = inc_weak = \
@@ -1272,7 +1236,7 @@ scanelf_match_symname(elfobj *elf, char *found_sym, char **ret, size_t *ret_len,
 		} else {
 			if (g_match) {
 				/* regex match the symbol */
-				if (rematch(this_sym, symname, REG_EXTENDED) != 0)
+				if (regexec(find_sym_regex_arr->eles[n], symname, 0, NULL, 0) == REG_NOMATCH)
 					continue;
 
 			} else if (*this_sym) {
@@ -1298,14 +1262,12 @@ scanelf_match_symname(elfobj *elf, char *found_sym, char **ret, size_t *ret_len,
 
 			goto matched;
 		}
-	} while (next_sym);
+	}
 
 	return;
 
  matched:
 	*found_sym = 1;
-	if (next_sym)
-		next_sym[-1] = saved;
 }
 
 static const char *scanelf_file_sym(elfobj *elf, char *found_sym)
@@ -1661,8 +1623,13 @@ static int scanelf_fileat(int dir_fd, const char *filename, const struct stat *s
 			return 1;
 	}
 	fd = openat(dir_fd, filename, (fix_elf ? O_RDWR : O_RDONLY) | O_CLOEXEC);
-	if (fd == -1)
+	if (fd == -1) {
+		if (fix_elf && errno == ETXTBSY)
+			warnp("%s: could not fix", filename);
+		else if (be_verbose > 2)
+			printf("%s: skipping file: %s\n", filename, strerror(errno));
 		return 1;
+	}
 
 	if (scanelf_elf(filename, fd, st->st_size) == 2) {
 		/* if it isn't an ELF, maybe it's an .a archive */
@@ -2021,7 +1988,7 @@ static const char * const opts_help[] = {
 	"Find a specified symbol",
 	"Find a specified section",
 	"Find a specified library",
-	"Use regex matching rather than string compare (use with -s)",
+	"Use regex rather than string compare (with -s); specify twice for case insensitive",
 	"Locate cause of TEXTREL",
 	"Print only ELF files matching etype ET_DYN,ET_EXEC ...",
 	"Print only ELF files matching numeric bits",
@@ -2127,8 +2094,14 @@ static int parseargs(int argc, char *argv[])
 			xarraypush_str(find_section_arr, optarg);
 			break;
 		case 's': {
-			if (find_sym) warn("You prob don't want to specify -s twice");
-			find_sym = optarg;
+			/* historically, this was comma delimited */
+			char *this_sym = strtok(optarg, ",");
+			if (!this_sym)	/* edge case: -s '' */
+				xarraypush_str(find_sym_arr, "");
+			while (this_sym) {
+				xarraypush_str(find_sym_arr, this_sym);
+				this_sym = strtok(NULL, ",");
+			}
 			break;
 		}
 		case 'N':
@@ -2183,7 +2156,7 @@ static int parseargs(int argc, char *argv[])
 			break;
 		}
 		case 'Z': show_size = 1; break;
-		case 'g': g_match = 1; break;
+		case 'g': ++g_match; break;
 		case 'L': load_cache_config = use_ldcache = 1; break;
 		case 'y': scan_symlink = 0; break;
 		case 'A': scan_archives = 1; break;
@@ -2225,7 +2198,43 @@ static int parseargs(int argc, char *argv[])
 	}
 	if (show_textrels && be_verbose)
 		has_objdump = bin_in_path("objdump");
+	/* precompile all the regexes */
+	if (g_match) {
+		regex_t preg;
+		const char *this_sym;
+		size_t n;
+		int flags = REG_EXTENDED | REG_NOSUB | (g_match > 1 ? REG_ICASE : 0);
+
+		array_for_each(find_sym_arr, n, this_sym) {
+			/* see scanelf_match_symname for logic info */
+			switch (this_sym[0]) {
+			case '%':
+				while (*(this_sym++))
+					if (*this_sym == '%') {
+						++this_sym;
+						break;
+					}
+				break;
+			case '+':
+			case '-':
+				++this_sym;
+				break;
+			}
+			if (*this_sym == '*')
+				++this_sym;
+
+			ret = regcomp(&preg, this_sym, flags);
+			if (ret) {
+				char err[256];
+				regerror(ret, &preg, err, sizeof(err));
+				err("regcomp of %s failed: %s", this_sym, err);
+			}
+			xarraypush(find_sym_regex_arr, &preg, sizeof(preg));
+		}
+	}
 	/* flatten arrays for display */
+	if (array_cnt(find_sym_arr))
+		find_sym = array_flatten_str(find_sym_arr);
 	if (array_cnt(find_lib_arr))
 		find_lib = array_flatten_str(find_lib_arr);
 	if (array_cnt(find_section_arr))
@@ -2317,15 +2326,27 @@ static int parseargs(int argc, char *argv[])
 		ret = scanelf_dir(search_path);
 	}
 
+#ifdef __PAX_UTILS_CLEANUP
 	/* clean up */
 	xarrayfree(ldpaths);
+	xarrayfree(find_sym_arr);
 	xarrayfree(find_lib_arr);
 	xarrayfree(find_section_arr);
+	free(find_sym);
 	free(find_lib);
 	free(find_section);
+	{
+		size_t n;
+		regex_t *preg;
+		array_for_each(find_sym_regex_arr, n, preg)
+			regfree(preg);
+		xarrayfree(find_sym_regex_arr);
+	}
 
 	if (ldcache != 0)
 		munmap(ldcache, ldcache_size);
+#endif
+
 	return ret;
 }
 
